

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath> 
#include <vector>
#include <complex> 
#include <string>
#include <ctime>  

extern "C" {
	// -------- LAPACK routines 
	// diagonalize symmetric real 
	void dsyev_(const char* choose, const char* uplow, long int* N, double* A, long int* rownum, double* Evals, double* WORK, long int* LWORK, long int* INFO );
	// diagonalize Hermitian 
	void zheev_(const char* choose, const char* uplow, long int* N, double* A, long int* rownum, double* Evals, double* WORK, long int* LWORK, double* RWORK, long int* INFO ); 
	// -------- BLAS routines 
	// matrix operating on a vector 
	void zgemv_(const char*, long int*, long int*, double*, double*, long int*, double*, long int*, double*, double*, long int*); 
	// matrix x matrix
	void zgemm_(const char*, const char*, long int*, long int*, long int*, double*, double*, long int*, double*, long int*, double*, double*, long int*); 
}

long int diagonalize(long int Dim, std::complex<double>* Mat, double* Evals);
long int diagonalize_real(long int Dim, double* Mat, double* Evals); 
long int tensor_prod_real(long int DimA, double* A, long int DimB, double* B, long int& DimC, double* C);
long int tprod_addto_real(long int DimA, double* A, long int DimB, double* B, double* C);
long int A_on_v(    long int N, std::complex<double>* v, std::complex<double>* A); 
long int AT_on_v(   long int N, std::complex<double>* v, std::complex<double>* A); 
long int A_times_B( long int N, std::complex<double>* C, std::complex<double>* A, std::complex<double>* B); 
long int AT_times_B(long int N, std::complex<double>* C, std::complex<double>* A, std::complex<double>* B); 
long int A_times_BT(long int N, std::complex<double>* C, std::complex<double>* A, std::complex<double>* B); 
long int trace_bath(long int Ntot, long int Nsys, std::complex<double>* psi, std::complex<double>* rho); 


int main(long int argc, char* argv[]){ 
	
//---------- switches 
	bool calc_eigs = true;
	bool simulate  = true; 
	
//---------- data filename and bath size 
	std::string fname = "20k_g002_s5e3"; 
	long int Bath_size = 5000; 
	std::string pops_fbase = "pops_d150at1450_"; 
	
//---------------------------------------initialize parameters

	const long int seed = 123456;        // random seed
	const double pi = 3.1415927;         // fundamental constants
	const std::complex<double> ei(0,1);
	
	long int Nsys = 4,
			 Ntrans = (Nsys*(Nsys-1))/2, 
			 Nbath = Bath_size,
			 Ntot = Nsys*Nbath,
			 Mat_size = Ntot*Ntot;  
	
	const double T = 2000, 
				 Dt = 10;      // time step    
					   
	const long int ns = static_cast<long int>(T/Dt); 
	
	const unsigned long int n_out = 50; 
	
	double beta = 0.4; 
	
	long int Nband = 150,         
			 start_band = 1450;  
				 
	const double E_min = 1,
				 E_max = 10,
				 DE = (E_max-E_min)/(Nbath-1), 
				 g = 0.01,        // sys/bath coupling strength 
				 spread = 0.005;         
				
//--------------------------------initialize random numbers 
	std::srand(seed);
	
//--------------------------------initialize timing 
	time_t time0, *time_ptr = 0;       
	time0 = time(time_ptr); 	
	
//--------------------------------data file extensions etc 
	std::string fxbin = fname + ".bin", fxdat = fname + ".dat";
	long int nblocks, bsize, block; 

//------------------ create the Hamiltonian 
	
	// pointers to matrices and vectors 
	double *UH, *Evs; 
	std::complex<double> *UHC; 
	
	//workhorse pointers
	double *ptr, *fin; 
	std::complex<double> *ztr, *zin, *ztr2; 
		
	// the system interaction X  
	double Xsys[16] = {   0, -0.7, 0.3,  0.9,
					   -0.7,    0, 1.2, -0.4, 
						0.3,  1.2,   0,  0.4, 
						0.9, -0.4, 0.4,    0 }; 
	double XEvals[Nsys]; 
	double XUsys_real[Nsys*Nsys]; for (long int n=0;n<Nsys*Nsys;n++) { XUsys_real[n] = Xsys[n]; } 
	diagonalize_real(Nsys, XUsys_real, XEvals); 
	// make Xsys_diag
	double Xsys_diag[Nsys*Nsys]; 
	for (long int n=0;n<Nsys*Nsys;n++) { Xsys_diag[n] = 0.0; } 
	for (long int n=0;n<Nsys;n++) { Xsys_diag[n*(Nsys+1)] = XEvals[n]; } 	
	// make complex copy of XUsys 
	std::complex<double> XUsys[Nsys*Nsys]; 
	for (long int n=0;n<Nsys*Nsys;n++) { XUsys[n] = XUsys_real[n]; } 
	
	// complex system Hamiltonian 
	double Evals[Nsys]; 
	Evals[0] = 0.5;  
	Evals[1] = 1.5;  
	Evals[2] = 2.2;  
	Evals[3] = 4.0;  
	std::complex<double> Hsys_z[Nsys*Nsys]; for (long int n=0;n<Nsys*Nsys;n++) { Hsys_z[n] = 0.0; }
	Hsys_z[0] = Evals[0]; Hsys_z[5] = Evals[1]; Hsys_z[10] = Evals[2]; Hsys_z[15] = Evals[3];
	// note: DX = XU' X XU
	std::complex<double> Temp[Nsys*Nsys]; 
	AT_times_B(Nsys,Temp,XUsys,Hsys_z); 
	A_times_B(Nsys,Hsys_z,Temp,XUsys); 
	
	// real system Hamiltonian 
	double Hsys[Nsys*Nsys]; for (long int n=0;n<Nsys*Nsys;n++) { Hsys[n] = Hsys_z[n].real(); } 
	
	// now Usys is the dagger of XUsys 
	// so: D = U' H U
	std::complex<double> Usys[Nsys*Nsys]; 
	for (long int n=0;n<Nsys;n++) { 
		for (long int m=0;m<Nsys;m++) { Usys[n + m*Nsys] = std::conj(XUsys[m + n*Nsys]); } 
	} 

	// make vector of transition energies 
	double Etrans[6]; ptr = Etrans; 
	for (long int n=0;n<Nsys;n++) { 
		for (long int m=n+1;m<Nsys;m++) { *(ptr++) = Evals[m] - Evals[n]; } 
	} 
	
	// create the bath Hamiltonian 
	double Boltz[Nbath], Bath_Spec[Nbath]; 
	if (beta>0.001) { 
		for (long int n=0;n<Nbath;n++) { Boltz[n] = std::exp(-beta*(E_min + n*DE)); } 
		for (long int n=0;n<Nbath;n++) { 
			Bath_Spec[n] = (Boltz[n]-Boltz[0])*(E_max-E_min)/(Boltz[Nbath-1]-Boltz[0]) + E_min; 
		}
	} else {
		for (long int n=0;n<Nbath;n++) { Bath_Spec[n] = E_min + n*DE; }		
	}
	
	if (calc_eigs) { 
		
		UH = new double[Mat_size]; 
		// now initialize the Hamiltonian (called UH)
		ptr = UH; fin = &UH[Mat_size-1]; fin++; 
		while (ptr!=fin) { *(ptr++) = 0.0; } 
		
		// now add the terms:  [note: bath is the fast index]
		// ---- Hsys (adding the tensor product with I_bath)   
		for (long int j=0;j<Nsys;j++) { 
			for (long int k=0;k<Nsys;k++) { 
				double val = Hsys[j + k*Nsys]; 
				for (long int m=0;m<Nbath;m++) { 
					UH[m + j*Nbath + Ntot*(m + k*Nbath)] += val; 
				} 		
			}
		} 
		 
		// ---- Hbath  (adding the tensor product with I_sys)
		for (long int n=0;n<Nsys;n++) { 
			for (long int m=0;m<Nbath;m++) { 
				UH[m + n*Nbath + Ntot*(m + n*Nbath)] += Bath_Spec[m]; 
			} 		
		} 

		// ---- Xsys*Bint  (the interaction operator)
		if (spread==0) { 			
			
			double rnum, Bint_nm, small; 
			for (long int n=0;n<Nsys;n++) { 
				for (long int m=n+1;m<Nbath;m++) { 
					rnum = 2*static_cast<double>(std::rand())/static_cast<double>(RAND_MAX) - 1; 
					small = static_cast<double>(std::rand())/static_cast<double>(RAND_MAX) - 0.5;
					if (small<0.2) { small = 0.0; } else { small = 1.0; }
					Bint_nm = small*g*(1 + 0.075*rnum); // Bint(m,p), and Bint(p,m)*
					
					// now we have the indices m and p, and we loop over the system to 
					// create the tensor product of the elements (m,p) and (p,m) with Xsys 
					for (long int j=0;j<Nsys;j++) { 
						UH[n + j*Nbath + Ntot*(m + j*Nbath)] = XEvals[j]*Bint_nm; 
						UH[m + j*Nbath + Ntot*(n + j*Nbath)] = XEvals[j]*Bint_nm; // conj if complex 
					}
					
				}
			}
			
		} else {
			
			double Et, upp, lowend, hihend, Bint_mp, rnum, small; 
			long int m, p; 
			for (long int n=0;n<Ntrans;n++) { 
				Et = Etrans[n]; 
				upp = Bath_Spec[Nbath-1] - Et + spread; 
				m = 0; 
				while (Bath_Spec[m] <= upp) {   // NB: Et must be > spread 
					p = m + 1; 
					lowend = Bath_Spec[m] + Et - spread; 
					hihend = Bath_Spec[m] + Et + spread; 
					while (Bath_Spec[p] < lowend) { p++; } 
					while (p < Nbath && (Bath_Spec[p] >= lowend) && (Bath_Spec[p] <= hihend)) {  
					
						rnum = 2*static_cast<double>(std::rand())/static_cast<double>(RAND_MAX) - 1; 
						Bint_mp = g*(1 + 0.075*rnum); // Bint(m,p), and Bint(p,m)*
						
						// now we have the indices m and p, and we loop over the system to 
						// create the tensor product of the elements (m,p) and (p,m) with Xsys 
						for (long int j=0;j<Nsys;j++) { 
							UH[m + j*Nbath + Ntot*(p + j*Nbath)] = XEvals[j]*Bint_mp; 
							UH[p + j*Nbath + Ntot*(m + j*Nbath)] = XEvals[j]*Bint_mp; // conj if complex 
						}
					
						p++; 
					} 
					m++; 
				}
			}
			
		}
		
		std::cout << "Creating H took " 
				  << difftime(time(time_ptr),time0)/60.0 << " mins" << std::endl; 
			
//------------------ diagonalize H 
		
		time0 = time(time_ptr); 	
		
		Evs = new double[Ntot]; 
		
		diagonalize_real(Ntot, UH, Evs);  
			
		std::cout << "Diagonalization took " 
				  << difftime(time(time_ptr),time0)/60.0 << " mins" << std::endl; 

		// save U Evs
		std::string UE_fbase = "UE_", UE_fname = UE_fbase + fxbin; 
		std::ofstream output_UE(UE_fname.c_str(), std::ios::out | std::ios::binary);
		nblocks = 10; bsize = Ntot*(Ntot/nblocks); block = bsize*sizeof(UH[0]);
		output_UE.write(reinterpret_cast<char*>(Evs), Ntot*sizeof(Evs[0]));
		ptr = UH; 
		for (long int j=0;j<nblocks;j++) {
			output_UE.write(reinterpret_cast<char*>(ptr), block);
			ptr = ptr + bsize; 
		}
		output_UE.close(); 
			 
	}  // end calc_eigs
	
	//------------------------------------------------------------ 
	//                        Simulation
	//------------------------------------------------------------ 	
		
	if (simulate) { 
		
		if (!calc_eigs) { 
			UH = new double[Mat_size]; 
			Evs = new double[Ntot]; 
			// load U Evs 
			std::string UE_fbase = "UE_", UE_fname = UE_fbase + fxbin; 
			std::ifstream input_UH(UE_fname.c_str(), std::ios::in | std::ios::binary);
			input_UH.read(reinterpret_cast<char*>(Evs), Ntot*sizeof(Evs[0]));
			long int nblocks = 10, bsize = Ntot*(Ntot/nblocks), block = bsize*sizeof(UH[0]); 
			ptr = UH; 
			for (long int j=0;j<nblocks;j++) { 
				input_UH.read(reinterpret_cast<char*>(ptr), block);
				ptr = ptr + bsize; 
			}
			input_UH.close(); 		
		}
		
		time0 = time(time_ptr);

		// define initial states 
		std::complex<double> psi_sys0[4] = {0, 0, 1, 0};
		A_on_v(Nsys,psi_sys0,Usys); 
		
		std::complex<double> psi_bath0[Nbath]; 
		double bnorm = 0, r1, r2; 
		for (long int n = start_band-1;n<start_band-1+Nband;n++) { 
			r1 = static_cast<double>(rand())/static_cast<double>(RAND_MAX);
			r2 = static_cast<double>(rand())/static_cast<double>(RAND_MAX);
			psi_bath0[n] = (1.0 + r1)*std::exp(ei*2.0*pi*r2); 
			bnorm = bnorm + std::abs(psi_bath0[n])*std::abs(psi_bath0[n]); 
		} 
		bnorm = std::sqrt(bnorm); 
		for (long int n = start_band-1;n<start_band-1+Nband;n++) { psi_bath0[n] = psi_bath0[n]/bnorm; } 

		// now take the tensor product 
		std::complex<double> psi[Ntot];
		for (long int n=0;n<Nsys;n++) { 
			for (long int m=0;m<Nbath;m++) { psi[m + n*Nbath] = psi_sys0[n]*psi_bath0[m]; } 
		}
			
		// storage for pops time series 
		double pops[Nsys*ns + 1]; 
		
		// system density matrix 
		std::complex<double> rho[Nsys*Nsys], Mtemp[Nsys*Nsys]; 
		trace_bath(Ntot,Nsys,psi,rho); 
		// note: D = Usys' Hsys Usys 
		AT_times_B(Nsys,Mtemp,Usys,rho); 
		A_times_B(Nsys,rho,Mtemp,Usys); 
		for (long int m=0;m<Nsys;m++) { pops[m] = rho[m + m*Nsys].real(); }
		
		// make a complex copy of UH 
		UHC  = new std::complex<double>[Mat_size];
		ztr = UHC; 
		ptr = UH; fin = &UH[Mat_size - 1]; fin++; 
		while (ptr!=fin) { *(ztr++) = *(ptr++); }
		
		// calculate the evolution phases 
		std::complex<double> eiD[Ntot]; 
		ztr = eiD; zin = &eiD[Ntot-1]; zin++; 
		ptr = Evs; 
		while (ztr!=zin) { *(ztr++) = std::exp(-ei*(*(ptr++))*Dt); }	
		
		// evolution 
		for (long int n=1;n<=ns;n++) { 
							
			// evolve 
			AT_on_v(Ntot,psi,UHC); 
			ztr = psi; zin = &psi[Ntot-1]; zin++; 
			ztr2 = eiD; 
			while (ztr!=zin) { *ztr = (*ztr)*(*ztr2++); ztr++; } 
			A_on_v(Ntot,psi,UHC); 
			
			// get system populations
			trace_bath(Ntot,Nsys,psi,rho); 
			AT_times_B(Nsys,Mtemp,Usys,rho); 
			A_times_B(Nsys,rho,Mtemp,Usys); 
			for (long int m=0;m<Nsys;m++) { pops[m + n*Nsys] = rho[m + m*Nsys].real(); }
			
		}
				
		// output the evolution of the populations 
		std::string pops_fname = pops_fbase + fxdat; 
		std::ofstream output_pops(pops_fname.c_str());
		ptr = pops; 
		for (long int n=0;n<ns;n++) { 
			for (long int m=0;m<Nsys;m++) { output_pops << *(ptr++) << " "; } 
			output_pops << std::endl; 
		}
		output_pops.close();
		
		// report time for evolution 
		double mtook = difftime(time(time_ptr),time0)/60.0;
		double htook = difftime(time(time_ptr),time0)/3600.0; 
		std::cout << "Evolving took " << mtook  << " mins "; 
		std::cout << "(" << htook  << ") hours" << std::endl;  
				
	} 
	
// -------------------------------- ending tasks 
	
	delete[] UHC; 
	delete[] Evs;  
	delete[] UH; 

	return 0;
}

// ------------------------------------- 
//         additional routines 
// -------------------------------------

long int diagonalize(long int Dim, std::complex<double>* Mat, double* Evals) { 
	
	// The unitary U comes back as Mat
	// If the input matrix is Mat = H, then 
	//        H = U diag(Evals) U^T
	//   and  H U = U diag(eig)     so that the columns of U are the eigenvectors
	
	long int info = 0;
	const char* do_vecs = "V", *up = "U";
	double work_dummy[1];
	double* rwork = new double[Dim*3-2]; 
	double* Mat_ptr = reinterpret_cast<double*>(Mat); 
	double* Ev_ptr = reinterpret_cast<double*>(Evals); 
	
	// query to see what the optimal work size is
	long int lwork = -1; 
	zheev_(do_vecs, up, &Dim, Mat_ptr, &Dim, Ev_ptr, work_dummy, &lwork, rwork, &info);
	
	lwork = static_cast<long int>(work_dummy[0]);
	double* work = new double[2*lwork]; 
	
	// calculate the eigenvalues and eigenvectors of A
	zheev_(do_vecs, up, &Dim, Mat_ptr, &Dim, Ev_ptr, work, &lwork, rwork, &info);
	
	delete[] work; 
	delete[] rwork; 
	
	return info;	
}

long int diagonalize_real(long int Dim, double* Mat, double* Evals) { 
	
	// The unitary U comes back as Mat
	// If the input matrix is Mat = H, then 
	//        H = U diag(Evals) U^T
	//   and  H U = U diag(eig)     so that the columns of U are the eigenvectors
	
	long int info = 0;
	const char* do_vecs = "V", *up = "U";
	double work_dummy[1];
	
	// query to see what the optimal work size is
	long int lwork = -1; 
	dsyev_(do_vecs, up, &Dim, Mat, &Dim, Evals, work_dummy, &lwork, &info);
	
	lwork = static_cast<long int>(work_dummy[0]);
	double* work = new double[lwork]; 
	
	// calculate the eigenvalues and eigenvectors of A
	dsyev_(do_vecs, up, &Dim, Mat, &Dim, Evals, work, &lwork, &info);
	
	delete[] work; 
	
	return info;	
}

long int tensor_prod_real(long int DimA, double* A, long int DimB, double* B, long int& DimC, double* C) { 
	// Note: B is the fast index 
	long int Dimtot = DimA*DimB; 
	DimC = Dimtot; 
	double *ptr, *fin; 
	for (long int k=0;k<DimA;k++) { 
		for (long int n=0;n<DimA;n++) {
			double A_nk = A[n + k*DimA];
			for (long int l=0;l<DimB;l++) { 
				ptr = &C[n*DimB + Dimtot*(l + k*DimB)]; fin = ptr + DimB;
				double* B_ptr = &B[l*DimB]; 
				while (ptr!=fin) { *(ptr++) = A_nk*(*(B_ptr++));  }
				//				for (long int m=0;m<DimB;m++) {
				//						C[m + n*DimB + Dimtot*(l + k*DimB)] += A[n + k*DimA]*B[m + l*DimB]; 
				//				} 	
			}
		}
	} 	
	return 0; 
}

long int tprod_addto_real(long int DimA, double* A, long int DimB, double* B, double* C) { 
	// Note: B is the fast index 
	double *ptr, *fin; 
	long int Dimtot = DimA*DimB; 
	for (long int k=0;k<DimA;k++) { 
		for (long int n=0;n<DimA;n++) {
			double A_nk = A[n + k*DimA];
			for (long int l=0;l<DimB;l++) { 
				ptr = &C[n*DimB + Dimtot*(l + k*DimB)]; fin = ptr + DimB;
				double* B_ptr = &B[l*DimB]; 
				while (ptr!=fin) { *(ptr++) += A_nk*(*(B_ptr++));  }   // add_to 
			}
		}
	} 	
	return 0; 
}

long int A_on_v(long int N, std::complex<double>* v, std::complex<double>* A) { 
	
	const char* Mat_op = "N"; 
	std::complex<double> alpha[1] = {1.0}, beta[1] = {0.0}; 
	double* alphav = reinterpret_cast<double*> (alpha); 
	double* betav  = reinterpret_cast<double*> (beta); 
	long int incx[1] = {1}, incy[1] = {1}; 
	std::complex<double> x[N]; 
	// copy v to x 
	std::complex<double> *p_x = x, *fin_x = &x[N-1], *p_v = v; fin_x++; 
	while (p_x!=fin_x) { *(p_x++) = *(p_v++); } 
	
	double* A_ptr = reinterpret_cast<double*> (A);
	double* v_ptr = reinterpret_cast<double*> (v); 
	double* x_ptr = reinterpret_cast<double*> (x);
	
	zgemv_(Mat_op, &N, &N, alphav, A_ptr, &N, x_ptr, incx, betav, v_ptr, incy); 
	return 0; 
}

long int AT_on_v(long int N, std::complex<double>* v, std::complex<double>* A) { 
	
	const char* Mat_op = "C"; 
	std::complex<double> alpha[1] = {1.0}, beta[1] = {0.0}; 
	double* alphav = reinterpret_cast<double*> (alpha); 
	double* betav  = reinterpret_cast<double*> (beta); 
	long int incx[1] = {1}, incy[1] = {1}; 
	std::complex<double> x[N]; 
	// copy v to x 
	std::complex<double> *p_x = x, *fin_x = &x[N-1], *p_v = v; fin_x++; 
	while (p_x!=fin_x) { *(p_x++) = *(p_v++); } 
	
	double* A_ptr = reinterpret_cast<double*> (A);
	double* v_ptr = reinterpret_cast<double*> (v); 
	double* x_ptr = reinterpret_cast<double*> (x);
	
	zgemv_(Mat_op, &N, &N, alphav, A_ptr, &N, x_ptr, incx, betav, v_ptr, incy); 
	return 0; 
}


long int A_times_B(long int N, std::complex<double>* C, std::complex<double>* A, std::complex<double>* B) { 
	
	const char *A_Mat_op = "N", *B_Mat_op = "N"; 
	std::complex<double> alpha[1] = {1.0}, beta[1] = {0.0}; 
	double* alphav = reinterpret_cast<double*> (alpha); 
	double* betav  = reinterpret_cast<double*> (beta); 
	
	double* A_ptr = reinterpret_cast<double*> (A);
	double* B_ptr = reinterpret_cast<double*> (B); 
	double* C_ptr = reinterpret_cast<double*> (C);
	
	zgemm_(A_Mat_op, B_Mat_op, &N, &N, &N, alphav, A_ptr, &N, B_ptr, &N, betav, C_ptr, &N); 
	return 0; 
}

long int A_times_BT(long int N, std::complex<double>* C, std::complex<double>* A, std::complex<double>* B) { 
	
	const char *A_Mat_op = "N", *B_Mat_op = "C"; 
	std::complex<double> alpha[1] = {1.0}, beta[1] = {0.0}; 
	double* alphav = reinterpret_cast<double*> (alpha); 
	double* betav  = reinterpret_cast<double*> (beta); 
	
	double* A_ptr = reinterpret_cast<double*> (A);
	double* B_ptr = reinterpret_cast<double*> (B); 
	double* C_ptr = reinterpret_cast<double*> (C);
	
	zgemm_(A_Mat_op, B_Mat_op, &N, &N, &N, alphav, A_ptr, &N, B_ptr, &N, betav, C_ptr, &N); 
	return 0; 
}

long int AT_times_B(long int N, std::complex<double>* C, std::complex<double>* A, std::complex<double>* B) { 
	
	const char *A_Mat_op = "C", *B_Mat_op = "N"; 
	std::complex<double> alpha[1] = {1.0}, beta[1] = {0.0}; 
	double* alphav = reinterpret_cast<double*> (alpha); 
	double* betav  = reinterpret_cast<double*> (beta); 
	
	double* A_ptr = reinterpret_cast<double*> (A);
	double* B_ptr = reinterpret_cast<double*> (B); 
	double* C_ptr = reinterpret_cast<double*> (C);
	
	zgemm_(A_Mat_op, B_Mat_op, &N, &N, &N, alphav, A_ptr, &N, B_ptr, &N, betav, C_ptr, &N); 
	return 0; 
}

long int trace_bath(long int Ntot, long int Nsys, std::complex<double>* psi, std::complex<double>* rho) { 
	
	long int Nbath = Ntot/Nsys; 
	std::complex<double> sum; 
	for (long int n=0;n<Nsys;n++) { 
		for (long int m=0;m<Nsys;m++) { 
			std::complex<double> *n_ztr = &psi[n*Nbath], *n_zin = &psi[(n+1)*Nbath-1]; n_zin++; 
			std::complex<double> *m_ztr = &psi[m*Nbath]; 
			sum = 0.0; 
			while (n_ztr!=n_zin) { sum += (*(m_ztr++))*std::conj(*(n_ztr++)); }
			rho[m + n*Nsys] = sum;
		}
	}
	
	return 0;  
}
